

// CNF converter for propositional logic, derived from the
// Geometric Formula transformer of Geo.


#include <fstream>


#include "statisticscollector.h"

#include "firstorder/nnf.h"
#include "firstorder/simplifications.h"
#include "firstorder/replacement.h"
#include "firstorder/factor.h"
#include "firstorder/geometric.h"


#include "firstorder/testing.h"


#include "parser/parser.h"
#include "parser/token.h"
#include "parser/errors.h"
#include "parser/tptp.h"

#include "simplification.h"

#include "varmatching.h"

#include <iostream>

#include "varset.h"

#include "search.h" 
#include "false_instance.h"

#include "rulesystem.h"
#include "model.h"



#include "varsubst.h"
#include "varatom.h" 
#include "element.h" 
#include "predicate.h"
#include "groundatom.h" 
#include "partial.h"
#include "application.h"


#include "predicate.stat"
#include "firstorder/function.stat"
#include "prolog/funcpred.stat"





struct parameters
{
   bool tptp_input;
   bool tptp_output;
      // tptp_output is not yet supported.

   bool nonempty;
      // Insist on non-empty models. With nonempty = false,
      // the formula forall x. p(x) -> exists x. p(x) is not provable.

   std::string inputfilename;
      // File from which input is read. If inputfilename = empty, input
      // will be read from standard input. 
   
   std::ifstream* inputstream;


   explicit parameters( const std::list< std::string > & pars )
      : tptp_input( false ),
        tptp_output( false ),
	nonempty( false ),
	inputfilename( "" ),
	inputstream(0)  
   {

      if( pars. size( ) == 0 )
      {
         std::cout << "ERROR: There should be at least one parameter\n\n";
	 exit(0);
      }

      std::list< std::string > :: const_iterator p = pars. begin( ); 
      ++ p;

      while( p != pars. end( ))
      {
         if( *p == "-tptp_input" )
         {
            tptp_input = true;
	    ++ p; 
            goto recognized;
         }

	 if( *p == "-tptp_output" )
         {
            tptp_output = true;
	    ++ p;
	    goto recognized;
         }

	 if( *p == "-nonempty" )
         {
            nonempty = true;
	    ++ p;
	    goto recognized;
         }

	 if( *p == "-inputfile" ) 
         {
            ++ p;
	    if( p == pars. end( ))
            {
               std::cout << "ERROR: expecting filename after -inputfile\n\n";
	       exit(0);
            }

	    if( inputfilename. size( ))
            {
               std::cout << "ERROR: multiple filenames\n\n";
	       exit(0);
            }

	    inputfilename = *(p++);
	    inputstream = new std::ifstream( inputfilename. c_str( ));
            if( ! ( *inputstream ))
            {
               std::cout << "ERROR: ";
	       std::cout << "could not open inputfile ";
	       std::cout << inputfilename<< "\n\n";
	       exit(0);
            }

            goto recognized;
         }
      
	 std::cout << "ERROR: unrecognized parameter: " << *(p++);
	 std::cout << "\n";
	 exit(0);

      recognized:
	 ;
      }
   }


   ~parameters( )
   {
      delete inputstream;
   }


};


std::ostream& operator << ( std::ostream& stream , const parameters& pars )
{
   stream << "Parameters:\n\n";

   stream << "tptp_input =     " << pars. tptp_input << "\n";
   stream << "tptp_output =    " << pars. tptp_output << "\n";
   stream << "nonempty =       " << pars. nonempty << "\n";
   stream << "inputfile =      " << ( pars. inputfilename. size( ) ?
	                               pars. inputfilename :
				       "(stdin)" ) << "\n";
   stream << "\n";
   return stream;
}




// Return true if formula if is completely propositional.   

bool ispropositional( firstorder::formula f )
{
   switch( f. getoperator( ))
   {
   case firstorder::fol_false:
   case firstorder::fol_true:
   case firstorder::fol_not:
   case firstorder::fol_and:
   case firstorder::fol_or:
   case firstorder::fol_implies:
   case firstorder::fol_equiv:
      for( std::list< firstorder::formula > :: const_iterator 
              p = f. subformulas_begin( );
              p != f. subformulas_end( );
              ++ p )
      {
         if( ! ispropositional( *p ))
            return false;
      }
      return true;
   
   case firstorder::fol_forall:
   case firstorder::fol_exists:
      return false;

   case firstorder::fol_equals:
      return false;

   case firstorder::fol_atom:
      if( f. subterms_begin( ) != f. subterms_end( ))
         return false;
      return true;
   }

   ASSERT( false ); exit(0); 
}


// Decompose a disjunction of (hopefully) literals into a list of
// literals.


void decomposedisjunction( firstorder::formula f, 
                           std::list< firstorder::formula > & li )
{
   if( f. getoperator( ) == firstorder::fol_false )
      return;

   if( f. getoperator( ) == firstorder::fol_or )
   {
      for( std::list< firstorder::formula > :: const_iterator
              p = f. subformulas_begin( );
              p != f. subformulas_end( );
              ++ p )
      {
         decomposedisjunction( *p, li );
      }
      return;
   }

   if( f. getoperator( ) == firstorder::fol_atom || 
       f. getoperator( ) == firstorder::fol_not ) 
   {
      li. push_back(f);
      return;
   }

   ASSERT( false );
   exit(0);
}
 
 
int main( unsigned int nrargs, char* ( args [] ) )
{

#if 0

   variable v0;
   variable v1 = v0; v1 ++ ;
   variable v2 = v1; v2 ++ ;
   variable v3 = v2; v3 ++ ;
   variable v4 = v3; v4 ++ ;
   variable v5 = v4; v5 ++ ;
   variable v6 = v5; v6 ++ ;
   variable v7 = v6; v7 ++ ;
   variable v8 = v7; v8 ++ ;
   variable v9 = v8; v9 ++ ;
   variable v10 = v9; v10 ++ ; 
     
   element e0;
   element e1 = e0; e1 ++ ;
   element e2 = e1; e2 ++ ;
   element e3 = e2; e3 ++ ;
   element e4 = e3; e4 ++ ;
   element e5 = e4; e5 ++ ;
   element e6 = e5; e6 ++ ;
   element e7 = e6; e7 ++ ;
   element e8 = e7; e8 ++ ;
   element e9 = e8; e9 ++ ; 

   predicate aa = predicate::predicate( "a", 1 );
   predicate bb = predicate::predicate( "b", 1 );
   predicate pp = predicate::predicate( "p", 3 );
   predicate qq = predicate::predicate( "q", 4 );
   predicate rr = predicate::predicate( "r", 2 );

   analysis::dependencies dep1;
   analysis::dependencies dep2;
 
   std::vector< variable > vvvv1;
   vvvv1. push_back( v0 );
   vvvv1. push_back( v1 );
   vvvv1. push_back( v2 ); 
   vvvv1. push_back( v4 );

   varatom va1 = varatom::varatom( qq, vvvv1 );

   dep1. addequation( 1, 0 );
   dep1. addequation( 3, 2 );
   dep1. addequation( 5, 4 ); 

   // dep2. addequation( 2, 2 );

   std::list< unsigned int > arg1;
   std::list< unsigned int > arg2;

   arg1. push_back(2);
   arg1. push_back(4);
  
   arg2. push_back(3);
   arg2. push_back(5);

   dep1. addskolem( 0, pp, arg1 );
   dep1. addskolem( 15, pp, arg1 );

   dep2. addskolem( 1, pp, arg2 );
 
   std::cout << dep1 << "\n";
   std::cout << dep2 << "\n";
   std::cout << dep1. implies( dep2 ) << "\n";
   return 0;
 
   std::list< unsigned int > arg;
   arg. push_back( 1 );
   arg. push_back( 1 );

   dep1. addskolem( 2, qq, arg );

   std::vector< variable > vvvv2;
   vvvv2. push_back( v2 );
   vvvv2. push_back( v3 ); 
   vvvv2. push_back( v1 );
   
   varatom va2 = varatom::varatom( pp, vvvv2 );

   std::cout << dep1 << "\n";
   std::cout << dep2 << "\n";

   std::cout << va2 << "\n";

   std::cout << analysis::dependencies::inherit( va1, dep1, va2 ) << "\n";

   std::cout << dep1. implies( dep2 ) << "\n\n";

   std::cout << dep2 << "\n";
   std::cout << dep1 << "\n"; 
   return 0; 
                                         
   dep1. reduce( );
   dep2. reduce( );

   std::cout << dep1 << "\n";
   std::cout << dep2 << "\n";

   std::cout << "\n\n";
   std::cout << analysis::dependencies::merging( dep1, dep2 ) << "\n\n";

   return 0;

   model mod;

   std::vector< element > arguments;
   arguments. push_back( e0 );

   std::vector< unsigned int > parents;

   mod. addatom( groundatom::groundatom( predicate::create, arguments ), 
		                         44, parents ); 
	       

   arguments. clear( );
   arguments. push_back( e1 );
   mod. addatom( groundatom::groundatom( predicate::create, arguments ),
		                         33, parents );


   arguments. clear( );
   arguments. push_back( e0 );
   arguments. push_back( e1 );

   mod. closure_counter ++ ;

   parents. push_back(0);
   parents. push_back(1);

   mod. addatom( groundatom::groundatom( pp, arguments ),
		 22, parents );

   parents. clear( );
   parents. push_back(2);
   parents. push_back(2);

   mod. addatom( groundatom::groundatom( rr, arguments ),
		 23, parents );

   std::cout << mod << "\n\n"; 
   mod. backtrack(2);
   std::cout << mod << "\n";

   return 0;
#endif 

   std::cout << "hallo\n\n";

   std::list< std::string > pppp;
   for( unsigned int i = 0; i < nrargs; ++ i )
   {
      pppp. push_back( args [i] );
   }

   parameters pars = parameters::parameters( pppp );

   std::cout << "\n";
   std::cout << pars << "\n";

   parser::inputstream is = 
      ( pars. inputfilename. size( ) ? 
	      parser::inputstream::inputstream( *( pars. inputstream ),
			                        pars. inputfilename ) :
	      parser::inputstream::inputstream( )
      );



   model m;
   rulesystem rs;
   firstorder::sequent fol;


   if( pars. tptp_input )
   {
      try
      {
         parser::tptp::readproblem( is, fol );
      }
      catch( parser::syntaxerror s )
      {
         std::cout << "ERROR: ";
	 std::cout << s << "\n";
	 return 0;
      }
      catch( parser::tptp::conversion_error c )
      {
         std::cout << "ERROR: ";
	 std::cout << c << "\n";
	 return 0;
      }
   }
   else
   {
      try
      {
         readproblem( is, m, rs, fol );
      }
      catch( parser::syntaxerror s )
      {
         std::cout << "ERROR: "; 
         std::cout << s << "\n\n";
         return 0; 
      }
   }

   fol. collectusedidentifiers( );

   std::cout << m << "\n";
   std::cout << rs << "\n"; 

   firstorder::simplifications::flatten( fol ); 

   if( fol. formulas. size( ))
   {

      firstorder::simplifications::flatten( fol );
      std::cout << "-----------------------------------------------------\n";
      std::cout << "Input:\n\n";
      std::cout << fol << "\n";


      firstorder::nnf::convert_to_nnf( fol );
      firstorder::simplifications::flatten( fol ); 
      std::cout << "-----------------------------------------------------\n"; 
      std::cout << "NNF:\n\n";
      std::cout << fol << "\n";


      firstorder::simplifications::fully_simplify( fol );
      std::cout << "-----------------------------------------------------\n";
      std::cout << "Full simplification:\n\n";
      std::cout << fol << "\n";


      for( std::list< firstorder::formula > :: const_iterator
              p = fol. formulas. begin( );
              p != fol. formulas. end( );
              ++ p )
      {
         if( ! ispropositional( *p ))
         {
            std::cout << "formula : " << *p;
            std::cout << " is not first-order\n\n";
            return 0;
         }
      }      

      firstorder::factor::convert_to_cuden( fol );
      firstorder::simplifications::flatten( fol );
      std::cout << "-----------------------------------------------------\n";
      std::cout << "Factored normal form (CUDEN):\n\n";
      std::cout << fol << "\n";


      std::cout << "-----------------------------------------------------\n";
      std::cout << "Clauses:\n\n";

      for( std::list< firstorder::formula > :: const_iterator
              p = fol. formulas. begin( );
              p != fol. formulas. end( );
              ++ p )
      {
         std::list< firstorder::formula > clause;
         decomposedisjunction(*p, clause );

         std::cout << "[ "; 
         for( std::list< firstorder::formula > :: const_iterator
                 p = clause. begin( );
                 p != clause. end( );
                 ++ p )
         {
            if( p != clause. begin( ))
               std::cout << ", ";

            if(  p -> getoperator( ) == firstorder::fol_atom )
               std::cout << *p;
            else
            {
               std::cout << "- " << *( p -> subformulas_begin( )); 
            }
         }
         std::cout << " ]\n";
      }
   }

   return 0; 
}




